import os
import cv2
import json
import numpy as np
from flow_viz import flow_to_image
import pycocotools.mask as mask_utils
from skimage.morphology import dilation
from skimage.segmentation import find_boundaries
from label_helper import color2id, id2label
import random

random.seed(100)
DEPTH_SCALE_MAX = 100.0
MAX_FLOW = 695
MIN_FLOW = -1687


def id2rgb(id):
    blue = id // (256 * 256)
    id %= 256 * 256
    green = id // 256
    id %= 256
    red = id
    return np.array([red, green, blue])


def read_flow(filename):
    flow_img = cv2.imread(filename, cv2.IMREAD_UNCHANGED)
    flow_img = flow_img[:, :, ::-1]
    flow_img = flow_img.astype(np.float32)
    flow_img = (flow_img[..., :2] / np.iinfo(np.uint16).max) * (MAX_FLOW - MIN_FLOW) + MIN_FLOW
    return flow_img


def read_depth_png(depth_png):
    depth = cv2.imread(depth_png, cv2.IMREAD_UNCHANGED)
    return (depth.astype(np.float32) / np.iinfo(np.uint16).max) * 1000.0


def color_depth(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        depth = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        not_valid_mask = depth == 0
        depth_scaled = (depth / float(np.iinfo(np.uint16).max)) * 1000
        depth_scaled = (depth_scaled / DEPTH_SCALE_MAX) * 255
        depth_scaled = depth_scaled.astype(np.uint8)
        depth_scaled = cv2.applyColorMap(depth_scaled, cv2.COLORMAP_JET)
        if overlay:
            img = cv2.imread(
                file_path.replace("depth", "images").replace("_depth.png", "_rgb.jpg")
            )
            depth_scaled = cv2.addWeighted(img, 0.5, depth_scaled, 0.5, 0)
        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), depth_scaled)


def color_flow(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        flow = read_flow(file_path)
        flow = flow_to_image(flow)
        if overlay:
            img = cv2.imread(
                file_path.replace("flow", "images").replace("_flow.png", "_rgb.jpg")
            )
            flow = cv2.addWeighted(img, 0.5, flow, 0.5, 0)
        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), flow)


def color_semseg(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        sem = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        color_sem = np.zeros((sem.shape[0], sem.shape[1], 3), dtype=np.uint8)
        for color, id in color2id.items():
            color_sem[np.where(sem == id)] = color[::-1]
        if overlay:
            img = cv2.imread(
                file_path.replace("semantic_seg", "images").replace(
                    "_seg.png", "_rgb.jpg"
                )
            )
            color_sem = cv2.addWeighted(img, 0.3, color_sem, 0.7, 0)

        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), color_sem)


def color_insseg(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        insem = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        mapping = json.load(open(file_path.replace(".png", ".json")))
        color_ins = np.zeros((insem.shape[0], insem.shape[1], 3), dtype=np.uint8)
        curr_inst_id = insem % 1000
        for id in np.unique(curr_inst_id):
            if id == 0:
                continue
            id_mask = curr_inst_id == id
            color_ins[id_mask] = id2rgb(mapping[str(id)])[::-1]
        if overlay:
            img = cv2.imread(
                file_path.replace("instance_seg", "images").replace(
                    "_inseg.png", "_rgb.jpg"
                )
            )
            color_ins = cv2.addWeighted(img, 0.3, color_ins, 0.7, 0)
        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), color_ins)


def color_panseg(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        pansem = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        mapping = json.load(open(file_path.replace(".png", ".json")))
        color_pan = np.zeros((pansem.shape[0], pansem.shape[1], 3), dtype=np.uint8)
        for id in np.unique(pansem):
            id_mask = pansem == id
            if id < 1000:
                color_pan[id_mask] = id2label[id].color[::-1]
            else:
                ins_id = id % 1000
                color_pan[id_mask] = id2rgb(mapping[str(ins_id)])[::-1]
        if overlay:
            img = cv2.imread(
                file_path.replace("panoptic_seg", "images").replace(
                    "_pano.png", "_rgb.jpg"
                )
            )
            color_pan = cv2.addWeighted(img, 0.3, color_pan, 0.7, 0)

        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), color_pan)


def color_amodal_panseg(
    source_dir, target_dir, file_lists, overlay=False, proc_id=None
):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        pansem = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        amodal_info = json.load(open(file_path.replace(".png", ".json")))
        colored_gt = np.zeros((pansem.shape[0], pansem.shape[1], 3), dtype=np.uint8)
        for id in np.unique(pansem):
            id_mask = pansem == id
            if id < 1000:
                colored_gt[id_mask] = id2label[id].color[::-1]
            else:
                color_id = amodal_info[str(id)]["track_id"]
                color = np.array(id2label[id // 1000].color)[::-1]
                occ_flag = amodal_info[str(id)]["occluded"]
                if occ_flag:
                    id_mask = mask_utils.decode(
                        amodal_info[str(id)]["amodal_mask"]
                    ).astype(np.bool)
                contours = (
                    find_boundaries(id_mask, mode="thick", background=0).astype(
                        np.uint8
                    )
                    * 255
                )
                id_mask[contours != 0] = 0
                colored_gt[id_mask] = colored_gt[id_mask] * 0.5 + color * 0.5
                colored_gt[contours != 0] = np.uint8(np.array([255, 255, 255]))
        if overlay:
            img = cv2.imread(
                file_path.replace("panoptic_seg", "images").replace(
                    "_pano.png", "_rgb.jpg"
                )
            )
            colored_gt = cv2.addWeighted(img, 0.3, colored_gt, 0.7, 0)
        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), colored_gt)


def color_amodal_insseg(
    source_dir, target_dir, file_lists, overlay=False, proc_id=None
):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        pansem = cv2.imread(file_path, cv2.IMREAD_UNCHANGED)
        amodal_info = json.load(open(file_path.replace(".png", ".json")))
        colored_gt = np.zeros((pansem.shape[0], pansem.shape[1], 3), dtype=np.uint8)
        for id in np.unique(pansem):
            id_mask = pansem == id
            if id < 1000:
                colored_gt[id_mask] = id2label[id].color[::-1]
            else:
                color_id = amodal_info[str(id)]["track_id"]
                color = id2rgb(color_id)[::-1]
                occ_flag = amodal_info[str(id)]["occluded"]
                if occ_flag:
                    id_mask = mask_utils.decode(
                        amodal_info[str(id)]["amodal_mask"]
                    ).astype(np.bool)
                contours = (
                    find_boundaries(id_mask, mode="thick", background=0).astype(
                        np.uint8
                    )
                    * 255
                )
                id_mask[contours != 0] = 0
                colored_gt[id_mask] = colored_gt[id_mask] * 0.5 + color * 0.5
                colored_gt[contours != 0] = np.uint8(np.array([255, 255, 255]))
        if overlay:
            img = cv2.imread(
                file_path.replace("panoptic_seg", "images").replace(
                    "_pano.png", "_rgb.jpg"
                )
            )
            colored_gt = cv2.addWeighted(img, 0.3, colored_gt, 0.7, 0)

        cv2.imwrite(os.path.join(target_dir, camera, seq_name, file_name), colored_gt)


def color_amodal_semseg(
    source_dir, target_dir, file_lists, overlay=False, proc_id=None
):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        files = os.listdir(file_path)
        files = [filename.decode("utf-8") for filename in files]
        sem_all = {}
        for file in files:
            if file.endswith(".png"):
                sem = cv2.imread(os.path.join(file_path, file), cv2.IMREAD_UNCHANGED)
                sem_all[file.split("_")[-2]] = sem
        save_file_name = "_".join(file.split("_")[:-2]) + "_{}".format(
            file.split("_")[-1]
        )
        color_sem = color_amodal_semantic(sem_all)
        cv2.imwrite(
            os.path.join(target_dir, camera, seq_name, save_file_name), color_sem
        )


def color_amodal_semantic(sem_all):
    ordered_keys = sorted(sem_all.keys())
    color_sem = np.zeros(
        (sem_all[ordered_keys[0]].shape[0], sem_all[ordered_keys[0]].shape[1], 3),
        dtype=np.uint8,
    )
    sem = sem_all[ordered_keys[0]]
    for color, id in color2id.items():
        color_sem[np.where(sem == id)] = np.array(color[::-1])
    if len(ordered_keys) > 1:
        for key in ordered_keys[1:]:
            sem = sem_all[key]
            if isinstance(key, str):
                key = int(key)
            color = id2label[key].color[::-1]
            mask = sem == key
            color_sem[mask] = color_sem[mask] * 0.2 + 0.8 * np.array(color)

    return color_sem


def color_amodal_depth(source_dir, target_dir, file_lists, overlay=False, proc_id=None):
    for file_path in file_lists:
        camera = file_path.split("/")[-3]
        seq_name = file_path.split("/")[-2]
        file_name = file_path.split("/")[-1]
        os.makedirs(os.path.join(target_dir, camera, seq_name), exist_ok=True)
        files = os.listdir(file_path)
        files = [filename.decode("utf-8") for filename in files]
        sem_all = {}
        for file in files:
            file = str(file)
            if file.endswith(".png"):
                file_ = file
                sem = read_depth_png(os.path.join(file_path, file))
                sem_all[file.split("_")[-2]] = sem
        save_file_name = "_".join(file_.split("_")[:-2]) + "_{}".format(
            file_.split("_")[-1]
        )
        modal_depth = color_depth_(
            read_depth_png(
                os.path.join(
                    ("/")
                    .join(file_path.split("/")[:-1])
                    .replace("amodal_depth", "depth"),
                    save_file_name.replace("_amdep.png", "_dep.png"),
                )
            )
        )
        color_sem = color_amodal_depth_helper(sem_all)
        mask = np.logical_and(
            np.logical_and(color_sem[:, :, 0] == 0, color_sem[:, :, 1] == 0),
            color_sem[:, :, 2] == 0,
        )
        color_sem[mask] = modal_depth[mask]

        cv2.imwrite(
            os.path.join(target_dir, camera, seq_name, save_file_name), color_sem
        )


def color_depth_(depth_scaled):
    depth_scaled = (depth_scaled / DEPTH_SCALE_MAX) * 255
    depth_scaled = depth_scaled.astype(np.uint8)
    depth_scaled = cv2.applyColorMap(depth_scaled, cv2.COLORMAP_JET)
    return depth_scaled


def color_amodal_depth_helper(sem_all):
    ordered_keys = sorted(sem_all.keys())
    color_sem = np.zeros(
        (sem_all[ordered_keys[0]].shape[0], sem_all[ordered_keys[0]].shape[1], 3),
        dtype=np.uint8,
    )
    sem = sem_all[ordered_keys[0]]
    valid_mask = sem != 0
    color_sem[valid_mask] = color_depth_(sem)[valid_mask]
    if len(ordered_keys) > 1:
        for key in ordered_keys[1:]:
            sem = sem_all[key]
            valid_mask = sem != 0
            color_sem[valid_mask] = (
                color_sem[valid_mask] * 0.5 + 0.5 * color_depth_(sem)[valid_mask]
            )

    return color_sem
