import argparse
import os
import multiprocessing
import numpy as np
import cv2
import glob

from color_utils import (
    color_semseg,
    color_insseg,
    color_panseg,
    color_depth,
    color_flow,
    color_amodal_panseg,
    color_amodal_semseg,
    color_amodal_depth,
    color_amodal_insseg,
)


def parse_args():
    parser = argparse.ArgumentParser(description="Reorganize the dataset")
    parser.add_argument(
        "--source_dir",
        type=str,
        default="/home/mohan/datasets/amodalSynthDrive/train/",
    )
    parser.add_argument(
        "--target_dir",
        type=str,
        default="/home/mohan/datasets/colored/train/",
    )
    parser.add_argument(
        "--type",
        nargs="+",
        default=[
            "amodal_semantic",
        ],
    )
    parser.add_argument(
        "--camera",
        nargs="+",
        default=[
            "back",
        ],
    )
    parser.add_argument("--cpu_count", type=int, default=32)
    args = parser.parse_args()
    return args


def multi_core(
    task_fn, source_dir, target_dir, file_lists, overlay=False, cpu_count=None
):
    sequence_names = file_lists

    if cpu_count is None:
        cpu_count = (
            len(sequence_names)
            if len(sequence_names) < multiprocessing.cpu_count()
            else multiprocessing.cpu_count()
        )
    sequences_split = np.array_split(sequence_names, cpu_count)
    print(
        "Number of cores: {}, sequences per core: {}".format(
            cpu_count, len(sequences_split[0])
        )
    )
    workers = multiprocessing.Pool(processes=cpu_count)
    processes = []
    for proc_id, sequence_set in enumerate(sequences_split):
        processes.append(
            workers.apply_async(
                task_fn, (source_dir, target_dir, sequence_set, overlay, proc_id)
            )
        )

    for proc in processes:
        proc.get()

    workers.close()


def main():
    args = parse_args()
    source_dir = args.source_dir
    target_dir = args.target_dir
    data_type = args.type
    cameras = args.camera
    cpu_count = args.cpu_count

    if not isinstance(data_type, list):
        data_type = [
            data_type,
        ]
    if not isinstance(cameras, list):
        cameras = [
            cameras,
        ]
    os.makedirs(target_dir, exist_ok=True)

    data_type_mapping = {
        "semantic": {"folder": "semantic_seg", "function": color_semseg},
        "instance": {"folder": "instance_seg", "function": color_insseg},
        "panoptic": {"folder": "panoptic_seg", "function": color_panseg},
        "depth": {"folder": "depth", "function": color_depth},
        "flow": {"folder": "flow", "function": color_flow},
        "amodal_panoptic": {
            "folder": "amodal_panoptic_seg",
            "function": color_amodal_panseg,
        },
        "amodal_instance": {
            "folder": "amodal_instance_seg",
            "function": color_amodal_insseg,
        },
        "amodal_semantic": {
            "folder": "amodal_semantic_seg",
            "function": color_amodal_semseg,
        },
        "amodal_depth": {"folder": "amodal_depth", "function": color_amodal_depth},
    }

    overlay = False  # add functionality to overlay the colorized gt on the original image later
    for type_i in data_type:
        save_dir = os.path.join(target_dir, "color_{}".format(type_i))
        file_dir = os.path.join(source_dir, data_type_mapping[type_i]["folder"])

        os.makedirs(save_dir, exist_ok=True)
        file_lists = []
        for camera in cameras:
            if type_i == "amodal_semantic" or type_i == "amodal_depth":
                file_list = glob.glob(os.path.join(file_dir, "{}/*/*".format(camera)))
            else:
                file_list = glob.glob(
                    os.path.join(file_dir, "{}/*/*.png".format(camera))
                )
            file_lists.extend(file_list)

        print("{}".format(len(file_lists)))
        multi_core(
            data_type_mapping[type_i]["function"],
            file_dir,
            save_dir,
            sorted(file_lists),
            overlay,
            cpu_count,
        )


if __name__ == "__main__":
    main()
